package no.rwr.awt.x;
import java.awt.*;
import no.rwr.x.GC;
import no.rwr.x.XImage;
import no.rwr.x.Drawable;
import no.rwr.x.Window;

public class XGraphics extends Graphics implements Cloneable {
  
  

    GC context;
    XGraphicsConfiguration config;
    Rectangle clipBounds;
    int xt, yt;
    
    Color fg;
    Color bg;
    
    XFontMetrics metrics;

    public void dispose() {
	context.dispose();
	context = null;
	config = null;
	clipBounds = null;
    }



    public XGraphics(Drawable drawable, XGraphicsConfiguration config,
		     Color bg) {
	context = new GC(drawable);
	this.config = config;
	this.bg = bg;
    }
    

    public Graphics create() {
	XGraphics gfx;
	try {
	    gfx = (XGraphics) clone();
	} catch (CloneNotSupportedException cnse) {
	    throw new Error("should never happen:" + cnse.toString());
	}
	gfx.context = context.create();

	return gfx;
    }
    
    public void translate(int x, int y) {
	xt += x;
	yt += y;
    }

    public Color getColor() {
	return fg;
    }
    
    
    public void setColor(Color color) {
	fg = color;
	setColorImpl(fg);
    }

    final void setColorImpl(Color color) {
	context.setForeground(config.getPixel(color));
    }

    public void setPaintMode() {
	throw new UnsupportedOperationException("not implemented");
    }

    public void setXORMode(Color c1) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    
    public Font getFont() {
	return metrics.getFont();
    }
    
    public void setFont(Font font) {
	if ((metrics != null) && font.equals(metrics.getFont())) return;
	metrics = config.getXFontMetrics(font);
	context.setFont(metrics.xfont);
    }
    
    public FontMetrics getFontMetrics(Font font) {
	if ((metrics != null) && font.equals(metrics.getFont())) {
	    return metrics;
	}

	return config.getXFontMetrics(font);
    }
    
    public Rectangle getClipBounds() {
	Rectangle clip = (Rectangle) clipBounds.clone();
	clip.translate(-xt, -yt);
	return clip; // FIXME: return clip shape not bounds
    }
    
    
    public void clipRect(int x, int y, int width, int height) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    
    public void setClip(int x, int y, int width, int height) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    
    public Shape getClip() {
	return getClipBounds();
    }
    
    
    public void setClip(Shape clip) {
	/* TODO: create a special RectangleUnion shape that can be
	   used to draw advantage of the GCs ability to set multiple
	   rectangles. 
	*/

	/* FIXME: creating all these objects is wasteful and can be
	   costly in the long run, since this code is run at every
	   expose. */
	Rectangle newClipBounds = clip.getBounds();
	newClipBounds.x += xt;
	newClipBounds.y += yt;
	if ((clipBounds != null) && !clipBounds.contains(newClipBounds)) {
	    System.err.println("warning: old clip ("+ clipBounds +") does " +
			       "not fully contain new clip (" +
			       newClipBounds + ")");
	}
	clipBounds = newClipBounds;
	Rectangle[] rects = { clipBounds };
	context.setClipRectangles(rects);
    }
    
    
    public void copyArea(int x, int y, int width, int height, int
			 dx, int dy) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    
    public void drawLine(int x1, int y1, int x2, int y2) {
	context.drawLine(x1+xt, y1+yt, x2+xt, y2+yt);
    }
    
    
    public void fillRect(int x, int y, int width, int height) {
	context.fillRectangle(x+xt, y+yt, width, height);
    }
    
    
    public void clearRect(int x, int y, int width, int height) {
	setColor(bg);
	fillRect(x, y, width, height);
	setColor(fg);

	/* We can't use clearArea here, since clear area does not
	 respect clipping.
	boolean exposures = false;
	context.clearArea(x+xt, y+yt, width, height, exposures);
	*/
    }
    
    
    public void drawRoundRect(int x, int y, int width, int
			      height, int arcWidth, int arcHeight) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    
    public void fillRoundRect(int x, int y, int width, int
			      height, int arcWidth, int arcHeight) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    
    public void drawOval(int x, int y, int width, int height) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    
    public void fillOval(int x, int y, int width, int height) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    
    public void drawArc(int x, int y, int width, int height, int
			startAngle, int arcAngle) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    
    public void fillArc(int x, int y, int width, int height, int
			startAngle, int arcAngle) {
	throw new UnsupportedOperationException("not implemented");
    }
    

    public void drawPolyline(int[] xPoints, int[] yPoints, int
			     nPoints) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    
    public void drawPolygon(int[] xPoints, int[] yPoints, int
			    nPoints) {
	throw new UnsupportedOperationException("not implemented");
    }
    


    public void fillPolygon(int[] xPoints, int[] yPoints, int
			    nPoints) {
	throw new UnsupportedOperationException("not implemented");
    }


    public void drawString(String str, int x, int y) {
	context.drawString(str, x+xt, y+yt);
    }


    /*
    public void drawString(AttributedCharacterIterator iterator,
			   int x, int y) {
	throw new UnsupportedOperationException("not implemented");
    }
    */

    public boolean drawImage(Image img, int x, int y,
			     ImageObserver observer) {
	if (!(img instanceof BufferedImage)) {
	    System.err.println("unknown image class");
	    return false;
	}
	
	BufferedImage bimg = (BufferedImage) img;

	XImage ximg = (XImage) bimg.getProperty("no.rwr.x.XImage");
	if (ximg == null) {
	    System.err.println("skipping null XImage");
	    // FIXME: do on the spot conversion

	    return false;
	}

	x+=xt;
	y+=yt;

	/*

	  +------------------
	  |    clip
	  |     +---------+
	  | img |         |
	  |  +--+-------+ |
	  |  |  |       | | 
	  |  |  |       | |
	  |  |  +-------+-+
	  |  |          |
	  |  +----------+

	*/


	int iLeft   = Math.max(x, clipBounds.x);
	int iTop    = Math.max(y, clipBounds.y);
	int iRight  = Math.min(x + bimg.getWidth(),
			       clipBounds.x + clipBounds.width);
	int iBottom = Math.min(y + bimg.getHeight(),
			       clipBounds.y + clipBounds.height);
	
	int srcX = iLeft - x;
	int srcY = iTop  - y;

	int width  = iRight  - iLeft;
	int height = iBottom - iTop;

	if ((width <= 0) || (height <= 0)) {
	    return true;
	}

	context.putImage(ximg, srcX, srcY, iLeft, iTop,
			 width, height);
	return true;
    }


    public boolean drawImage(Image img, int x, int y, int width,
			     int height, ImageObserver observer) {
	throw new UnsupportedOperationException("not implemented");
    }

    public boolean drawImage(Image img, int x, int y, Color
			     bgcolor, ImageObserver observer) {
	throw new UnsupportedOperationException("not implemented");
    }


    public boolean drawImage(Image img, int x, int y, int width,
			     int height, Color bgcolor,
			     ImageObserver observer) {
	throw new UnsupportedOperationException("not implemented");
    }


    public boolean drawImage(Image img, int dx1, int dy1, int dx2, int dy2,
			     int sx1, int sy1, int sx2, int sy2,
			     ImageObserver observer) {
	throw new UnsupportedOperationException("not implemented");
    }


    public boolean drawImage(Image img, int dx1, int dy1, int dx2, int dy2,
			     int sx1, int sy1, int sx2, int sy2,
			     Color bgcolor, ImageObserver observer) {
	throw new UnsupportedOperationException("not implemented");
    }

}
